package de.heinerkuecker.iterator.primitive;

import java.util.ConcurrentModificationException;
import java.util.NoSuchElementException;

public interface IntIterator
{
	boolean hasNext();
	
	int next();

	@Deprecated // no modification counter
	public static final IntIterator EMPTY =
			new IntIterator()
			{
				@Override
				public boolean hasNext()
				{
					return false;
				}
				
				@Override
				public int next()
				{
					throw new NoSuchElementException();
				}
			};

	/**
	 * Get empty int iterator.
	 * 
	 * @param modificationCountbl
	 * @return empty {@link IntIterator} with modification counter
	 */
	public static IntIterator empty(
			final ModificationCountbl modificationCountbl )
	{
		return 
				new IntIterator()
				{
					private final int modificationCounter = modificationCountbl.modificationCounter();
		
					@Override
					public boolean hasNext()
					{
						if ( this.modificationCounter != modificationCountbl.modificationCounter() )
						{
							throw new ConcurrentModificationException();
						}
						return false;
					}
					
					@Override
					public int next()
					{
						if ( this.modificationCounter != modificationCountbl.modificationCounter() )
						{
							throw new ConcurrentModificationException();
						}
						throw new NoSuchElementException();
					}
				};
	}
}
